@@ -1,5 +1,7 @@
 from __future__ import annotations
 
+import functools
+
 import dataclasses
 import hashlib
 import json
@@ -9,7 +11,7 @@ import sys
 import threading
 import typing as T
 from collections import namedtuple
-from functools import cached_property, wraps
+from functools import cached_property
 from pathlib import Path
 from typing import Generic, TypeVar, Dict
 
@@ -19,7 +21,9 @@ import numpy as np
 import tiktoken
 from tqdm import tqdm
 
+from codebased.gitignore import parse_gitignore
 from codebased.settings import EmbeddingsConfig
+from codebased.utils import decode_text
 
 if T.TYPE_CHECKING:
     from openai import OpenAI
@@ -60,10 +64,10 @@ class Events:
     StoreEmbeddings = namedtuple('StoreEmbeddings', ['embeddings'])
     Commit = namedtuple('Commit', [])
     FaissInserts = namedtuple('FaissInserts', ['embeddings'])
-    IndexObjects = namedtuple('IndexObjects', ['file_bytes', 'objects'])
+    IndexObjects = namedtuple('IndexObjects', ['file_bytes', 'objects', 'lines'])
     ScheduleEmbeddingRequests = namedtuple('EmbeddingRequests', ['requests'])
     FaissDeletes = namedtuple('FaissDeletes', ['ids'])
-    IndexFile = namedtuple('IndexFile', ['path', 'content'])  # tuple[Path, bytes]
+    IndexFile = namedtuple('IndexFile', ['path', 'content', 'lines'])
     DeleteFile = namedtuple('DeleteFile', ['path'])
     DeleteFileObjects = namedtuple('DeleteFile', ['path'])
     Directory = namedtuple('Directory', ['path'])
@@ -76,27 +80,6 @@ def is_binary(file_bytes: bytes) -> bool:
     return b'\x00' in file_bytes
 
 
-def is_utf8(file_bytes: bytes) -> bool:
-    try:
-        file_bytes.decode('utf-8')
-        return True
-    except UnicodeDecodeError:
-        return False
-
-
-def is_utf16(file_bytes: bytes) -> bool:
-    # Check for UTF-16 BOM (Byte Order Mark)
-    if file_bytes.startswith(b'\xff\xfe') or file_bytes.startswith(b'\xfe\xff'):
-        return True
-
-    # If no BOM, check if the file decodes as UTF-16
-    try:
-        file_bytes.decode('utf-16')
-        return True
-    except UnicodeDecodeError:
-        return False
-
-
 # Put this on Dependencies object.
 class OpenAIRequestScheduler:
     def __init__(self, oai_client: "OpenAI", embedding_config: EmbeddingsConfig):
@@ -202,26 +185,31 @@ class ThreadSafeCache(Generic[K, V]):
 class thread_local_cached_property:
     def __init__(self, func):
         self.func = func
-        self.local = threading.local()
+        self.name = None
+
+    def __set_name__(self, owner, name):
+        self.name = name
 
     def __get__(self, obj, cls=None):
         if obj is None:
             return self
 
-        if not hasattr(self.local, 'instance'):
-            self.local.instance = self.func(obj)
-        return self.local.instance
+        thread_id = threading.get_ident()
+        attr_name = f'_thread_local_cache_{self.name}_{thread_id}'
 
-    def __set_name__(self, owner, name):
-        self.name = name
+        if not hasattr(obj, attr_name):
+            setattr(obj, attr_name, self.func(obj))
+        return getattr(obj, attr_name)
 
     def clear_cache(self, obj):
-        if hasattr(self.local, 'instance'):
-            delattr(self.local, 'instance')
+        thread_id = threading.get_ident()
+        attr_name = f'_thread_local_cache_{self.name}_{thread_id}'
+        if hasattr(obj, attr_name):
+            delattr(obj, attr_name)
 
 
 def clear_thread_local_cache(func):
-    @wraps(func)
+    @functools.wraps(func)
     def wrapper(self, *args, **kwargs):
         result = func(self, *args, **kwargs)
         for name, attr in type(self).__dict__.items():
@@ -238,6 +226,7 @@ class Dependencies:
     config: Config
     settings: Settings
     search_cache: ThreadSafeCache[Flags, list] = dataclasses.field(default_factory=ThreadSafeCache)
+    ignores: dict[Path, T.Callable[[Path], bool | None]] = dataclasses.field(default_factory=dict)
 
     @cached_property
     def openai_client(self) -> "OpenAI":
@@ -271,6 +260,20 @@ class Dependencies:
         except FileNotFoundError:
             return lambda _: False
 
+    def ignore_checker(self, path: Path) -> bool:
+        original_path = path
+        while path.is_relative_to(self.config.root):
+            try:
+                ignore_rule_set = self.ignores[path / '.gitignore']
+                ignore_result = ignore_rule_set(original_path)
+                if ignore_result is not None:
+                    return ignore_result
+            except KeyError:
+                continue
+            finally:
+                path = path.parent
+        return False
+
     @cached_property
     def request_scheduler(self) -> OpenAIRequestScheduler:
         return OpenAIRequestScheduler(self.openai_client, self.settings.embeddings)
@@ -303,7 +306,7 @@ def index_paths(
         *,
         total: bool = True
 ):
-    ignore = dependencies.gitignore
+    ignore = dependencies.ignore_checker
     db = dependencies.db
     index = dependencies.index
 
@@ -341,7 +344,19 @@ def index_paths(
                 path = event.path
                 if path == config.root / '.git' or path == config.root / '.codebased':
                     continue
-                for entry in os.scandir(path):
+                dir_entries = list(os.scandir(path))
+                # TODO: We don't handle changes to .gitignore files in the background worker
+                #  because it could require re-scanning the entire index.
+                #  i.e. if you remove an ignore rule, we might need to add previously ignored files to the index.
+                #  or if you add a new ignore rule, we might need to remove existing files from the index.
+                try:
+                    gitignore_file = next(e for e in dir_entries if e.name == '.gitignore' and e.is_file())
+                    ignore_path = Path(gitignore_file.path)
+                    gitignore_parsed = parse_gitignore(ignore_path, base_dir=path)
+                    dependencies.ignores[ignore_path] = gitignore_parsed
+                except StopIteration:
+                    pass
+                for entry in dir_entries:
                     entry_path = Path(entry.path)
                     if ignore(entry_path):  # noqa
                         continue
@@ -394,7 +409,8 @@ def index_paths(
                         raise FileExceptions.Ignore()
                     # TODO: See how long this takes on large repos.
                     # TODO: We might want to memoize the "skip" results if this is an issue.
-                    if not (is_utf8(file_bytes) or is_utf16(file_bytes)):
+                    decoded_text = decode_text(file_bytes)
+                    if decoded_text is None:
                         raise FileExceptions.Ignore()
                     real_sha256_digest = hashlib.sha256(file_bytes).digest()
                     # TODO: To support incremental indexing, i.e. allowing this loop to make progress if interrupted
@@ -422,7 +438,7 @@ def index_paths(
                     if previous_sha256_digest == real_sha256_digest:
                         raise FileExceptions.AlreadyIndexed()
                     # Actually schedule the file for indexing.
-                    events.append(Events.IndexFile(relative_path, file_bytes))
+                    events.append(Events.IndexFile(relative_path, file_bytes, decoded_text.splitlines()))
                     # Delete old objects before adding new ones.
                     events.append(Events.DeleteFileObjects(relative_path))
                     continue
@@ -503,7 +519,7 @@ def index_paths(
                     # )
                 deletion_markers.extend(deleted_ids)
             elif isinstance(event, Events.IndexFile):
-                relative_path, file_bytes = event.path, event.content
+                relative_path, file_bytes, lines = event.path, event.content, event.lines
                 assert isinstance(relative_path, Path)
                 assert isinstance(file_bytes, bytes)
 
@@ -531,17 +547,16 @@ def index_paths(
 
                     ).fetchone()
                     objects_by_id[object_id] = obj
-                events.append(Events.IndexObjects(file_bytes, objects_by_id))
+                events.append(Events.IndexObjects(file_bytes, objects_by_id, lines))
             elif isinstance(event, Events.IndexObjects):
                 file_bytes = event.file_bytes
+                in_lines = event.lines
                 objects_by_id = event.objects
                 # dict[int, Object]
                 assert isinstance(objects_by_id, dict)
-                # vector stuff!
-                lines = file_bytes.split(b'\n')
                 requests_to_schedule = []
                 for obj_id, obj in objects_by_id.items():
-                    rendered = render_object(obj, in_lines=lines, file=False)
+                    rendered = render_object(obj, in_lines=in_lines, file=False)
                     if not rendered:
                         continue
                     request = EmbeddingRequest(
